/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (c) 2000-2001 Silicon Graphics, Inc.  All rights reserved.
 */
#ifndef _ASM_IA64_SN_MMZONE_TEMP_H
#define _ASM_IA64_SN_MMZONE_TEMP_H

#include <linux/config.h>
#include <linux/init.h>
#include <asm/processor.h>

#if defined(CONFIG_IA64_SGI_SN1)
#include <asm/sn/sn1/mmzone_sn1.h>
#elif defined(CONFIG_IA64_SGI_SN2)
#include <asm/sn/sn2/mmzone_sn2.h>
#elif defined(CONFIG_IA64_DIG)
#include <asm/mmzone_dig_numa.h>
#else
#error "Unknown architecture"
#endif


/*
 * General Concepts:
 *
 *	- Nodes are numbered several ways:
 *
 *		compact node numbers - compact node numbers are a dense numbering of 
 *			all the nodes in the system. An N node system will have compact
 *			nodes numbered 0 .. N-1. There is no significance to the node
 *			numbers. The compact node number assigned to a specific physical
 *			node may vary from boot to boot. The boot node is not necessarily
 *			node 0.
 *
 *		physical node numbers - Physical node numbers may not be dense
 *			nor do they necessarily start with 0. The exact significance of 
 *			a physical node number is platform specific.
 *
 *		proximity domain numbers - these numbers are assigned by ACPI. 
 *			Each platform must provide a platform specific function
 *			for mapping proximity node numbers to physical node numbers.
 *
 *	  Most of the code in the kernel uses compact node numbers to identify nodes.
 *
 *
 *	- Memory is conceptually divided into chunks. A chunk is either
 *	  completely present, or else the kernel assumes it is completely
 *	  absent. Each node consists of a number of possibly discontiguous chunks.
 * 
 *	- A contiguous group of memory chunks that reside on the same node
 *	  are referred to as a clump. Note that a clump may be partially present.
 *	  (Note, on some hardware implementations, a clump is the same as a memory
 *	  bank or a DIMM).
 *
 *      - a node consists of multiple clumps of memory. From a NUMA perspective,
 *        accesses to all clumps on the node have the same latency. Except for zone issues,
 *        the clumps are treated as equivalent for allocation/performance purposes.
 *
 *      - each node has a single contiguous mem_map array. The array contains page struct
 *        entries for every page on the node. There are no "holes" in the mem_map array.
 *	  The node directory (see below) has pointers to the start of the mem_map entries
 *	  for each clump on the node.
 *
 *      - associated with each node is a pg_data_t structure. This structure contains the
 *	  information used by the linux memory allocator for managing the memory on the
 *	  node. The pg_data_t structure for a node is located on the node.
 *
 *	- to minimize offnode memory references, a "node directory" is maintained on each
 *	  node. This directory replicates frequently used read-only data structures that
 *	  are used in macro evaluation. Examples include the addresses of the
 *	  pernode pg_data structures for each node.
 *
 *      - the MAP_NR function has been modified to be "clump aware" & uses the clump_mem_map_base
 *        array in the node directory for generating MAP_NR numbers.
 * 
 *	- the node directory contains array of pointers to the mem_map entries for each clump
 *	  of memory. The array is indexed by a platform specific function.
 *
 *	- each cpu has a pointer the node directory contained in it's cpu_data structure.
 *
 *	- each platform is responsible for defining the following constants & functions:
 *
 *		PLAT_BOOTMEM_ALLOC_GOAL(cnode,kaddr) - Calculate a "goal" value to be passed 
 *			to __alloc_bootmem_node for allocating structures on nodes so that 
 *			they dont alias to the same line in the cache as the previous 
 *			allocated structure. You can return 0 if your platform doesnt have
 *			this problem.
 *				(Note: need better solution but works for now ZZZ).
 *			
 *		PLAT_CHUNKNUM(kaddr) - takes a kaddr & returns its chunk number
 *
 *		PLAT_CLUMP_MEM_MAP_INDEX(kaddr) - Given a kaddr, find the index into the 
 *			clump_mem_map_base array of the page struct entry for the first page 
 *			of the clump.
 *
 *		PLAT_CLUMP_OFFSET(kaddr) - find the byte offset of a kaddr within the clump that
 *			contains it.
 *
 *		PLAT_CLUMPSIZE - defines the size in bytes of the smallest clump supported on the platform.
 *
 *		PLAT_CLUMPS_PER_NODE - maximum number of clumps per node
 *
 *		PLAT_MAXCLUMPS - maximum number of clumps on all node combined
 *
 *		PLAT_MAX_COMPACT_NODES - maximum number of nodes in a system. (do not confuse this
 *			with the maximum node number. Nodes can be sparsely numbered).
 *
 *		PLAT_MAX_NODE_NUMBER - maximum physical node number plus 1
 *
 *		PLAT_MAX_PHYS_MEMORY - maximum physical memory address
 *
 *		PLAT_PXM_TO_PHYS_NODE_NUMBER(pxm) - convert a proximity_domain number (from ACPI)
 *			into a physical node number
 *
 *		PLAT_VALID_MEM_KADDR(kaddr) - tests a kaddr to see if it potentially represents a
 *			valid physical memory address.  Return 1 if potentially valid, 0 otherwise.
 *			(This function generally tests to see if any invalid bits are set in
 *			the address).
 *
 *
 *	- each platform is responsible for defining the following typedefs::
 *	
 *		cnodeid_t	- compact node number
 *	
 */


#if 0	/* need new discontig.c for this */
extern struct page 	*invalid_mem_map;	/* value returned by virt_to_page for bad addresses */
#endif


/*
 * Node Directory. 
 * 
 * The  node directory is a read-only (after boot) structure
 * containing pointers to VM structures located on all nodes. The directory is
 * replicated on each node in order to reduce off-node references.
 */

typedef struct ia64_node_directory_s {
        pg_data_t	*pg_data_ptrs[PLAT_MAX_COMPACT_NODES];
	cnodeid_t	physical_node_map[PLAT_MAX_NODE_NUMBER];
	struct page	*clump_mem_map_base[PLAT_MAXCLUMPS];
} ia64_node_directory_t;


#define NODEDIR			((ia64_node_directory_t*)(local_cpu_data->node_directory))



/*
 * Given a compact node id, return a pointer to the pg_data_t for the node.
 * The following 2 macros are similar. 
 *
 *	NODE_DATA 	- should be used in all code not related to system
 *			  initialization. It uses pernode data structures to minimize
 *			  offnode memory references. However, these structure are not 
 *			  present during boot. This macro can be used once cpu_init
 *			  completes.
 *
 *	BOOT_NODE_DATA	- should be used during system initialization 
 *			  prior to freeing __initdata. It does not depend on the percpu
 *			  area being present.
 *
 */
#define NODE_DATA(nid)			(NODEDIR->pg_data_ptrs[nid])
#define BOOT_NODE_DATA(nid)		boot_get_pg_data_ptr((long)(nid))
extern pg_data_t* __init boot_get_pg_data_ptr(long);



/*
 * Chunk related macros
 *	Note: It is not clear if VALIDCHUNK is really needed. It is currently used
 *		ONLY in kern_address_valid. The non-NUMA variant of this always
 *		returns 1.
 *	ZZZ Fixme????
 */
#define VALIDCHUNK(cnum)		1



/*
 * Given a kaddr, find the base mem_map address for the start of the mem_map
 * entries for the clump containing the kaddr.
 */
#define CLUMP_MEM_MAP_BASE(kaddr)	NODEDIR->clump_mem_map_base[PLAT_CLUMP_MEM_MAP_INDEX(kaddr)]



/*
 * Given a kaddr, this macro return the relative map number 
 * within the clump.
 */
#define CLUMP_MAP_NR(kaddr) 		(PLAT_CLUMP_OFFSET(kaddr) >> PAGE_SHIFT)



/*
 * Finally.... This is the MAP_NR function for the platform.
 */
#define MAP_NR_DISCONTIG(kaddr)		({long _kmns=(long)(kaddr);				\
						CLUMP_MAP_NR(_kmns) +				\
						CLUMP_MEM_MAP_BASE(_kmns) - mem_map;})

/*
 * Given a pte, this macro returns a pointer to the page struct for the pte.
 */
#define pte_page(pte)			virt_to_page(PAGE_OFFSET | (pte_val(pte)&_PFN_MASK))



/*
 * Determine if a kaddr is a valid memory address of memory that
 * actually exists. 
 *
 * The check consists of 2 parts:
 *	- verify that the address is a region 7 address & does not 
 *	  contain any bits that preclude it from being a valid platform
 *	  memory address
 *	- verify that the chunk actually exists.
 *
 * Note that IO addresses are NOT considered valid addresses.
 *
 * Note, many platforms can simply check if kaddr exceeds a specific size.  
 *	(However, this wont work on SGI platforms since IO space is embedded 
 * 	within the range of valid memory addresses & nodes have holes in the 
 *	address range between clumps). 
 */
#define kern_addr_valid(kaddr)		({long _kav=(long)(kaddr);						\
					PLAT_VALID_MEM_KADDR(_kav) && VALIDCHUNK(PLAT_CHUNKNUM(_kav));})


/*
 * Given a kaddr, return a pointer to the page struct for the page.
 * If the kaddr does not represent RAM memory that potentially exists, return
 * a pointer the page struct for max_mapnr. IO addresses will
 * return the page for max_nr. Addresses in unpopulated RAM banks may
 * return undefined results OR may panic the system.
 *
 */
#if 0	/* need new discontig.c for invalid_mem_map */
#define virt_to_page(kaddr)		({long _kvtp=(long)(kaddr);					\
					(PLAT_VALID_MEM_KADDR(_kvtp))					\
						? CLUMP_MEM_MAP_BASE(_kvtp) + CLUMP_MAP_NR(_kvtp)	\
						: invalid_mem_map;})
#else
#define virt_to_page(kaddr)	({long _kvtp=(long)(kaddr);		\
				(PLAT_VALID_MEM_KADDR(_kvtp))		\
					? CLUMP_MEM_MAP_BASE(_kvtp) + CLUMP_MAP_NR(_kvtp)	\
					: (void*)0;})
#endif

/*
 * Given a page struct entry, return the physical address that the page struct represents.
 * Since IA64 has all memory in the DMA zone, the following works:
 */
#define page_to_phys(page)		__pa(page_address(page))


/*
 * phys_node_to_cnodeid - convert a physical node number  to a cnodeid
 */
#define phys_node_to_cnodeid(pnode)	(NODEDIR->physical_node_map[pnode])


/*
 * phys_node_present(pnode) - test a physical node number to see if it is present
 */
#define phys_node_present(pnode)	(NODEDIR->physical_node_map[pnode] != (cnodeid_t) -1)

#endif /* _ASM_IA64_SN_MMZONE_TEMP_H */
